
using System;
using System.Globalization;
using System.Text;

using Microsoft.Cci;
using Microsoft.Tools.FxCop.Sdk;
using Microsoft.Tools.FxCop.Sdk.Introspection;


namespace FxCop.Rules
{
    /// <summary>
    ///        <para>
    ///            Provides <see langword="static"/> helper methods for rules.
    ///        </para>
    /// </summary>
    internal /* static */ class RuleHelper
    {
        internal const string COMPANY_NAME = "MyCompany";        // Replace with your company name
        internal static readonly char[] DotCharArray = new char[] { '.' };
        internal static readonly string[] InternalNamespaces = new string[] { "native", "interop", "internal" };
        internal static readonly string[] InteropClasses = new string[] { "UnsafeInterop", "Interop", "SafeInterop", "NativeMethods", "UnsafeNativeMethods", "SafeNativeMethods" };

        private RuleHelper()
        {
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating if the specified type has a member of the specified type with the specified name.
        ///        </para>
        /// </summary>
        /// <param name="type">
        ///        The <see cref="TypeNode"/> to check.
        /// </param>
        /// <param name="memberType">
        ///        One of the <see cref="NodeType"/> representing the member to find.
        /// </param>
        /// <param name="name">
        ///        A <see cref="String"/> containing the name of the member to find.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if <paramref name="type"/> contains a member of type <paramref name="nodeType"/> and with the specified <paramref name="name"/>; otherwise, <see langword="false"/>.
        /// </returns>
        public static bool HasMemberWithName(TypeNode type, NodeType memberType, string name)
        {
            MemberList members = type.Members;

            for (int i = 0; i < members.Length; i++)
            {
                Member member = members[i];

                if (member.NodeType == memberType && member.Name.Name == name)
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating whether the specified name contains underscores.
        ///        </para>
        /// </summary>
        /// <param name="name">
        ///        A <see cref="String"/> containing the name to check.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if <paramref name="name"/> contains underscores; otherwise, <see langword="false"/>.
        /// </returns>
        /// <remarks>
        ///        <para>
        ///            For example, for 'backcolor', 'backColor', 'BackColor' <see cref="HasUnderscores"/> returns <see langword="false"/>, whereas, '_Backcolor', 'BACK_COLOR' and 'back_color' <see cref="HasUnderscores"/> returns <see langword="true"/>.
        ///        </para>
        /// </remarks>
        public static bool HasUnderscores(string name)
        {
            if (name == null)
                return false;

            return name.IndexOf('_') != -1;
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating whether specified name is Pascal-cased.
        ///        </para>
        /// </summary>
        /// <param name="name">
        ///        A <see cref="String"/> containing the name to check.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if <paramref name="name"/> is Pascal-cased otherwise, <see langword="false"/>.
        /// </returns>
        /// <remarks>
        ///        <para>
        ///            For example, for 'textBox', 'TEXTBOX', '_textbox' <see cref="IsPascal"/> returns <see langword="false"/>, whereas, '_TextBox' and 'TextBox' <see cref="IsPascal"/> returns <see langword="true"/>.
        ///        </para>
        /// </remarks>
        public static bool IsPascal(string name)
        {
            if (name == null || name.Length == 0)
                return true;

            string[] tokens = RuleUtilities.TokenizeCompoundWord(name);

            if (tokens.Length > 0 && RuleUtilities.CasingExceptions.ContainsKey(tokens[0]))
                return true;

            for (int i = 0; (i < name.Length); i++)
            {
                if (name[i] != '_')
                {
                    if (!char.IsUpper(name[i]))
                    {
                        return false;
                    }

                    break;
                }
            }

            return true;
        }


        /// <summary>
        ///        <para>
        ///            Replaces escape sequence characters in the specified text with a its textual equivalent.
        ///        </para>
        /// </summary>
        /// <param name="text">
        ///        A <see cref="String"/>.
        /// </param>
        /// <returns>
        ///        The <see cref="String"/> with its escape sequence characters replaced with its textual equivalent.
        /// </returns>
        public static string ReplaceEscapeCharacters(string text)
        {
            if (text == null)
                return null;

            StringBuilder builder = new StringBuilder(text.Length);

            for (int i = 0; i < text.Length; i++)
            {
                char c = text[i];

                switch (c)
                {
                    case '\n':
                        builder.Append(@"\n");
                        continue;

                    case '\r':
                        builder.Append(@"\r");
                        continue;

                    case '\a':
                        builder.Append(@"\a");
                        continue;

                    case '\b':
                        builder.Append(@"\b");
                        continue;

                    case '\f':
                        builder.Append(@"\f");
                        continue;

                    case '\t':
                        builder.Append(@"\t");
                        continue;

                    case '\v':
                        builder.Append(@"\v");
                        continue;

                    case '\0':
                        builder.Append(@"\0");
                        continue;
                }

                builder.Append(c);
            }

            return builder.ToString();
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating whether the specified class is an interops class.
        ///        </para>
        /// </summary>
        /// <param name="c">
        ///        The <see cref="Class"/> to check.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if <paramref name="method"/> is a interops class; otherwise, <see langword="false"/>.
        /// </returns>
        /// <remarks>
        ///        <para>
        ///            A class is considered an interops class if it one of the following names: Interop, UnsafeInterop, SafeInterop, NativeMethods, SafeNativeMethods or UnsafeNativeMethods.
        ///        </para>
        /// </remarks>
        /// <exception cref="ArgumentNullException">
        ///        <paramref name="c"/> is <see langword="null"/>.
        /// </exception>
        public static bool IsInteropClass(Class c)
        {
            if (c == null)
                throw new ArgumentNullException("c");

            string typeName = GetLastToken(c.Name.Name);

            foreach (string name in InteropClasses)
            {
                if (name == typeName)
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating whether the specified method is a platform invoke method.
        ///        </para>
        /// </summary>
        /// <param name="method">
        ///        The <see cref="Method"/> to check.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if <paramref name="method"/> is a platform invoke method; otherwise, <see langword="false"/>.
        /// </returns>
        /// <exception cref="ArgumentNullException">
        ///        <paramref name="method"/> is <see langword="null"/>.
        /// </exception>
        public static bool IsPInvoke(Method method)
        {
            if (method == null)
                throw new ArgumentNullException("method");

            return (method.Flags & MethodFlags.PInvokeImpl) == MethodFlags.PInvokeImpl;
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating whether the specified namespace name is an internal name.
        ///        </para>
        /// </summary>
        /// <param name="namespaceName">
        ///        A <see cref="String"/> containing the name of the namespace to check.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if the <paramref name="namespaceName"/> is an internal name; otherwise, <see langword="false"/>.
        /// </returns>
        /// <remarks>
        ///        <para>
        ///            An namespace is considered internal if its tokens contain any of the following words: Internal, Interop or Native.
        ///        </para>
        /// </remarks>
        public static bool IsInternalNamespace(string namespaceName)
        {
            return NamespaceContainsWord(namespaceName, InternalNamespaces);
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating whether the specified namespace name contains the specified words.
        ///        </para>
        /// </summary>
        /// <param name="namespaceName">
        ///        A <see cref="String"/> containing the name of the namespace to check.
        /// </param>
        /// <param name="words">
        ///        A <see cref="String"/> array containing the words.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if <paramref name="namespaceName"/> contains one or more <paramref name="words"/>; otherwise, <see langword="false"/>.
        /// </returns>
        public static bool NamespaceContainsWord(string namespaceName, string[] words)
        {
            if (namespaceName == null || namespaceName.Length == 0)
                return false;

            if (words == null || words.Length == 0)
                return false;

            string[] tokens = namespaceName.Split(DotCharArray);

            foreach (string token in tokens)
            {
                foreach (string word in words)
                {
                    if (string.Compare(token, word, true, CultureInfo.InvariantCulture) == 0)
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        /// <summary>
        ///        <para>
        ///            Returns the last token of the specified namespace name.
        ///        </para>
        /// </summary>
        /// <param name="namespaceName">
        ///        A <see cref="String"/> containing the name of the namespace.
        /// </param>
        /// <returns>
        ///        A <see cref="String"/> containing the last token of <see paramref="namespaceName"/> or <see langword="null"/> if <paramref name="namespaceName"/>.
        /// </returns>
        /// <remarks>
        ///        <para>
        ///            For example, for 'System.Windows.Forms', <see cref="GetNamespaceRoot"/> returns 'Forms'.
        ///        </para>
        /// </remarks>
        public static string GetLastToken(string namespaceName)
        {
            if (namespaceName == null || namespaceName.Length == 0)
                return namespaceName;

            int index = namespaceName.LastIndexOf('.');

            if (index == -1)
                return namespaceName;

            if (index + 1 == namespaceName.Length)
                return string.Empty;

            return namespaceName.Substring(index + 1);
        }

        /// <summary>
        ///        <para>
        ///            Returns the root namespace name of the specified namespace.
        ///        </para>
        /// </summary>
        /// <param name="namespaceName">
        ///        A <see cref="String"/> containing the name of the namespaces.
        /// </param>
        /// <returns>
        ///        A <see cref="String"/> containing the root namespace name of <paramref name="namespaceName"/> or <see langword="null"/> if <paramref name="namespaceName"/>.
        /// </returns>
        /// <remarks>
        ///        <para>
        ///            For example, for 'System.Windows.Forms', <see cref="GetNamespaceRoot"/> returns 'System'.
        ///        </para>
        /// </remarks>
        public static string GetNamespaceRoot(string namespaceName)
        {
            if (namespaceName == null || namespaceName.Length == 0)
                return namespaceName;

            int index = namespaceName.IndexOf('.');

            if (index == -1)
                return namespaceName;

            return namespaceName.Substring(0, index);
        }

        /// <summary>
        ///        <para>
        ///            Returns a value indicating whether the specified instruction is a method call.
        ///        </para>
        /// </summary>
        /// <param name="instruction">
        ///        The <see cref="Instruction"/> to check.
        /// </param>
        /// <returns>
        ///        <see langword="true"/> if <paramref name="instruction"/> is a method call; otherwise, <see langword="false"/>.
        /// </returns>
        /// <exception cref="ArgumentNullException">
        ///        <paramref name="instruction"/> is <see langword="null"/>.
        /// </exception>
        public static bool IsMethodCall(Instruction instruction)
        {
            if (instruction == null)
                throw new ArgumentNullException("instruction");

            switch (instruction.OpCode)
            {
                case OpCode.Callvirt:
                case OpCode.Call:
                case OpCode.Calli:
                    return true;
            }

            return false;
        }
    }
}

